******************************************************************************
; bmp converter - coverts PC 24 bitmaps into QL mode 32 and 33 bitmaps
; and saves these bitmaps into a file, as an image of the size of the
* current screen size, fit for loading with BGIMAGE
; (c) W. Lenerz 2002
; loosely based on PH. Dokos' win program
; v. 1.01 2002 Apr 13 06:59:29
; syntax:
*	BMPCVT32	file_in$,file_out$[,xsize%,ysize%[,red%,green%,blue%]]
*	BMPCVT33	file_in$,file_out$[,xsize%,ysize%[,red%,green%,blue%]]
* where:
*		file_in$	= PC Bitmap file
*		file_out$ = resulting QL file
*		xsize%,ysize% = QL image size, optional
*		  if these two parameters are omitted, then the image will
*		  be made to fit the current screen size
*		red%,green%,blue% = background colour if PC image < Ql image
*		the PC image is then centered
*******************************************************************************

	section cvt

	xdef	WL_bmpcvt32
	xdef	wl_bmpcvt33

; some macros

XJSR	macro	name
	xref	[name]		; - makes external jump
	JSR	[name]
	endm

Dtrap	macro	nbr,name
	trap	#[nbr]		; trap with error check
	tst.l	d0
	bne	[name]
	endm

; in ql screen (mode 32)  the colour bits are : gggbbbbb rrrrrggg
; in ql screen (mode 33)  the colour bits are : gggggrrr rrbbbbbw
; where w = intensity. set to 1 if at least two of the 6th bits (in the pc screen)
; for rgb are 1
; in pc 24 colour bitmap, the bits are bgr, and the lines are bottom down!
;

space	equ	$200
width	equ	$12		; x size of file is here (pixels)
height	equ	$16		; y size of file is here (pixels)
type	equ	$1c		; type of bitmap, must be 24 for 24 bit image
hlen	equ	$36



wl_bmpcvt32
	bsr.s	inp
	bsr	bord32
	bsr	got_size
	bsr	oloop32
	bsr	bord5		; this returns to basic!


wl_bmpcvt33
	bsr.s	inp
	bsr	bord33
	bsr	got_size
	bsr	oloop33
	bsr	bord5		; this returns to basic!



; first get the input filename, open file & load image into mem
inp	move.l	#space,d0	; get string from basic, into space
	xjsr	cpy_str 	; pointed then by a0
	bne	bmp_out 	; oops!
	move.l	a0,a4		; keep pointer to filename space
	moveq	#1,d3		; shared access
	moveq	#1,d0
	moveq	#-1,d1
	dtrap	2,bmp_out	; open file
	move.l	a4,a1
	move.l	#space,d2
	moveq	#$47,d0
	moveq	#-1,d3
	dtrap	3,bmp_out3	; read file header
	move.l	(a4),d5 	; length of file
	exg	a0,a4
	xjsr	gu_rchp 	; release mem for filename
	move.l	d5,d0		; length of file
	add.l	#space,d0	; + some space for myself
	xjsr	gu_achp 	; get mem for the file
	bne	bmp_out4		; oops->...
	exg	a0,a4		; a0 = channel id, a4 = memptr
	lea	space(a4),a1	; point to space for file
	move.l	d5,d2
	moveq	#$48,d0 	; greek goddess (scatter load file)
	dtrap	3,bmp_out3
	moveq	#2,d0
	trap	#2		; close input file now

; get & store the next parameter(s)
	addq.l	#8,a3		; point to next param
	move.l	a4,a0		; keep mem pointer
	move.l	#space-12,d0	; max length of output filename
	move.l	$58(a6),a1	; point maths stack
	xjsr	get_str 	; get string now (result filename)
	bne	bmp_out 	; oops!->...
	addq.l	#8,a3		; point to next param(s)
	cmp.l	a3,a5
	beq.s	get_size		; there are none ->..
	move.w	$112,a2
	jsr	(a2)		; get ints
	tst.l	d0		; errors?
	bne	bmp_out 	; yes ->...
	move.l	a4,a2		; point to filename
	move.w	(a2)+,d0		; length
	beq	err_bp		; what, 0 length filename????
	addq.w	#1,d0
	bclr	#0,d0		; make even
	add.w	d0,a2		; point after name
	move.w	d3,d0		; nbr of parameters got - keep
	beq.s	get_size		; there weren't any (?) ->...
	cmp.w	#2,d3		; there may be 2 ints...
	beq.s	param_ok
	cmp.w	#5,d3
	bne	err_bp		; ... or 5
param_ok	move.w	d3,(a2)+		; store nbr of params passed
	bra.s	do_cpy
cpy_lp	move.w	(a6,a1.l),(a2)+ ; possible x size,	y size, colours
	addq.l	#2,a1
do_cpy	dbf	d3,cpy_lp
	lsl.w	#1,d0		;
	add.w	d0,$58(a6)	; reset maths stack

; now calculate the image sizes, and how we must copy : 3 cases:
; pc image size = ql image size : no pb
; pc image size > ql image size : pc image will be cropped at both edges, so
;			    that it remains centered
* pc image size < ql image size : a border, possibly of [red,green,blue] will
;			    surround the image, an equal border is added
;			    to both sides (& up/down)
;			    if no colours were given, the border is black
get_size	clr.l	d1
	clr.l	d2
	clr.l	d4		;
	clr.l	d5		;
	clr.l	d6		;
	moveq	#-15,d0 	; error bp if wrong file type
	cmp.b	#$18,type+space(a4) ; correct file type?
	bne	bmp_out 	; no->...
	move.w	width+space(a4),d7	; image x size (nbr pix)in intel format
	beq	bmp_out 	; what, 0 sized image???? ->...
	rol.w	#8,d7		; now in comprehensible format
	move.w	d7,d1		; keep
	swap	d7
	move.w	height+space(a4),d7 ; image y size - nbr of lines
	beq	bmp_out 	; what, 0 sized image???? ->...
	rol.w	#8,d7		; now d7 = pc xsize | ysize
	move.w	d7,d0		; d0 = pc ysize
	move.l	d7,d6		: keep
	move.l	a4,a1
	move.w	(a1)+,d2		; name of length
	addq.w	#1,d2
	bclr	#0,d2		; make even
	add.l	d2,a1		; point to my parameters
	move.l	a1,a3
	move.w	(a1)+,d2		; are there any?
	beq.s	gs_out		; no->...
	move.w	(a1)+,d0		; ql xsize
	move.w	d0,d6
	swap	d6
	move.w	(a1)+,d1		; ql ysize
	move.w	d1,d6		; d6 = ql x | y size
gs_out	rts


; now make possible border - if none was given these should all be 0 (black)
; border for bmpcvt32
bord32	move.w	(a1)+,d5		; red
	andi.b	#%11111000,d5	; mask out lower 3 bits
	lsl.w	#8,d5		; into upper 5 bits
	move.w	(a1)+,d2		; green
	lsr.b	#2,d2		; mask out lower 2 bits
	lsl.w	#5,d2		; and move up
	or.w	d2,d5
	move.w	(a1)+,d2		; blue
	lsr.b	#3,d2		; mask out lower 3 bits
	or.b	d2,d5		; now we have rrrrrggggggbbbbb
	rol.w	#8,d5		; now we have gggbbbbb rrrrrggg
	clr.l	d3
	rts

; d5.w = border
; d6 = ql x | y size
; d7 = pc x | y size


got_size	move.l	d6,d2		; keep size
	mulu	d1,d0		; xsize * y size...
	lsl.l	#1,d0		; ... *2 = nbr of bytes needed for ql screen
	move.l	d0,(a3) 	; keep
	xjsr	gu_achp 	; get mem
	bne	bmp_out 	; ooops ->...
	move.l	a0,a5		; keep ql screen pointer

; now calculate possible cropping/border in the vertical
	move.w	d7,d0		; pc y size
	sub.w	d6,d0		; - ql y size
	beq.s	y_ok		; there is no difference ->...
	bgt.s	ql_smaly		; ql image is smaller
; here ql image is taller than pc image - we need a border above/below
	neg.w	d0
	lsr.w	#1,d0			; size of border on each side!
	move.w	d7,d6			; nbr of lines to copy from pc image (all of it)
	move.w	d0,d4			; d4 = nbr of lines for upper/lower border
	bra.s	y_ok
; here ql image is smaller than pc image - we need to crop

ql_smaly	lsr.w	#1,d0		; pixels to crop on each side
	move.w	d0,d4
	swap	d4			; d4 = nbr of lines to crop

; now calculate possible horizontal border/cropping
y_ok	swap	d6
	swap	d7
	move.w	d7,d0			; pc x size
	sub.w	d6,d0			; is there any size difference?
	beq.s	x_ok			; no ->...
	bgt.s	ql_smalx		; yes, ql image is smaller
; here ql image is larger, we need a border at left/right
	neg.w	d0
	lsr.w	#1,d0
	move.w	d0,d3			; nbr of pixels for border
	move.w	d7,d6			; nbr of bytes to copy from pc image
	bra.s	x_ok
ql_smalx
	lsr.w	#1,d0
	move.w	d0,d3			; nbr of pixels to crop on each side
	swap	d3

; here
; a4 = pc image
; a5 = ql image
; d2 = ql x | y size
; d3 : upper word<> 0 if ql image (x size) < pc image - cropping needed
; d3 : lower word <>0 if ql image > pc image - border needed
; d4 : upper word<> 0 if ql image (y size) < pc image - cropping needed
; d4 : lower word <>0 if ql image > pc image - border needed
; d5.w = border colour
; d6 = y | x size of image to copy
; d7 = y | x size of pc image
 
; now look at which line in the pc image we need to start. this will be either
; at the last line, or any line before that if we need to crop
x_ok	move.l	a5,a3			; point to my screen data
	lea	hlen+space(a4),a1	; point pc screen data
	clr.l	d0
	clr.l	d1
	move.w	d7,d0			; xsize of pc image
	mulu	#3,d0			; 3 bytes/pixel = bytes per line
	move.l	d0,-(a7)		; keep
	swap	d7
	move.w	d7,d1			; nbr of lines in pc image
	subq.w	#1,d1
	mulu	d1,d0
	add.l	d0,a1			; point to data for last line in pc image
	move.l	(a7)+,d1
	swap	d5
	move.w	d1,d5
	swap	d5			; d5 = bytes/line (pc) | border colour
	swap	d4			; .w <> if we need to crop
	mulu	d4,d1			; nbr of lines to crop * bytes/line
	sub.l	d1,a1			; point to 1st line to copy!
	clr.w	d4
	swap	d4			; d4 = possible nbr of hor. border lines
	beq.s	no_bord1		; there are none->...
; here we do the first border above/below the image
	move.w	d4,d1			; nbr of lines in border
	subq.w	#1,d1			; prepare for dbf
	swap	d2			; nbr of x pixels in ql image
	subq.w	#1,d2
brd_lp2 move.w	d2,d0		; nbr of x pixels in ql image
brd_lp1 move.w	d5,(a3)+		; copy border into screen
	dbf	d0,brd_lp1	; for each pixel in line
	dbf	d1,brd_lp2	; for each line
	swap	d4
	move.w	d2,d4		; keep this info for later
; if ql image (x size) < pc image, get offset into line
no_bord1	swap	d3		;
	clr.l	d0
	move.w	d3,d0		; nbr of pixels to crop from pc image
	mulu	#3,d0		; * nbr of bytes / pixel
	add.l	d0,a1		; point to start pixel in line
	clr.w	d3
	swap	d3		; nbr of pixels in border (if any)


; now we can copy each line from the image, possibly adding a border for each line
; here
; a1 = start of line to copy from
; a3 = current pixel in ql image
; a4 = pc image
; a5 = ql image
; d2 =
; d3 .w = nbr of pixels in x border
; d4 : nbr of lines in border (y dim) | nbr x pixel in ql image
; d5.w = bytes/line (pc)| border colour
; d6 = y | x size of image to copy
; d7 =
 

	subq.w	#1,d6		; nbr of pixels/line to copy
	swap	d6
	subq.w	#1,d6		; nbr of lines
	subq.w	#1,d3		; prepare for dbf
	rts

; this is the outer loop - cover each line (row) to treat
oloop32 swap	d6		; d6 = y | x size of image
	move.w	d6,d7		; nbr of pixels to copy
	swap	d6		; back to nbr of lines
	move.l	a1,a2		; point to 1rst byte of this line
	move.w	d3,d0		; any border at left hand side?
	blt.s	no_cpy		; no, so start converting pixels->...
ilp_cp1 move.w	d5,(a3)+		; copy left border for this line
	dbf	d0,ilp_cp1
no_cpy	clr.w	d0

; this is the inner loop:  covers each pixel in a line (row)
iloop	move.b	(a2)+,d0		; blue
	lsr.b	#3,d0			; get rid of the 3 lsbs
	move.b	(a2)+,d1		; green - upper bits go with red, lower with blue
	rol.b	#3,d1			; wrap it around
	move.b	d1,d2			; keep
	and.b	#%11100000,d2		; isolate 3 green bits
	or.b	d2,d0			; and make gggbbbbb
	lsl.w	#8,d0			; move into upper byte
	and.b	#%00000111,d1		; upper 3 bits for green
	move.b	(a2)+,d0
	and.b	#%11111000,d0
	or.b	d1,d0
	move.w	d0,(a3)+
	dbf	d7,iloop		; treat 1 line
; end of inner loop
	move.w	d3,d0			; any border at right hand side?
	blt.s	no_bord2		; no->...
ilp_cp2 move.w	d5,(a3)+		; copy left border for this line
	dbf	d0,ilp_cp2
no_bord2	swap	d5
	clr.l	d0
	move.w	d5,d0
;	 sub.w	 d5,a1		 ; point to beginning of next line
	sub.l	d0,a1		; point to beginning of next line
	swap	d5
	dbf	d6,oloop32	; do loop
	rts

; here we do the second border above/below the image
bord5	move.w	d4,d2
	swap	d4
	move.w	d4,d1		; nbr of lines in border
	beq.s	no_bord3		; there is no border->...
	subq.w	#1,d1		; prepare for dbf
brd_lp3 move.w	d2,d0		; nbr of x pixels in ql image
brd_lp4 move.w	d5,(a3)+		; copy border into screen
	dbf	d0,brd_lp4	; for each pixel in line
	dbf	d1,brd_lp3	; for each line

; now save the resulting file
no_bord3	move.l	a4,a0		; point to filename

	moveq	#1,d0
	moveq	#-1,d1
	moveq	#3,d3		; open overwrite...
	dtrap	2,bmp_out2
	move.l	a4,a1
	move.w	(a1)+,d0
	addq.w	#1,d0
	bclr	#0,d0
	add.w	d0,a1
	move.l	(a1),d2
	move.l	a5,a1		; where to start saving
	moveq	#$49,d0
	moveq	#-1,d3
	trap	#3		; greek goddess
	moveq	#2,d0
	trap	#2


bmp_out2	move.l	a5,a0		; release ql screen mem
	xjsr	gu_rchp
bmp_out move.l	a4,a0		; remease pc screen mem
	xjsr	gu_rchp
bmp_ok	addq.l	#4,a7		; skip return addresss (!)
	tst.l	d0
	rts			; return with(out) error

err_bp	moveq	#-15,d0 	; oops
	bra.s	bmp_out

bmp_out3	moveq	#2,d0		; close file
	trap	#2
	bra.s	bmp_out
bmp_out4	move.l	d0,d6
	moveq	#2,d0		; close file
	trap	#2
	addq.l	#4,a7		; jump over return address
	move.l	d6,d0
	rts

; routines for mode 33

; now make possible border - if none was given these should all be 0 (black)
; border for bmpcvt33 -  result must be :  gggggrrr rrbbbbbw
bord33	clr.l	d3
	clr.w	d2
	move.w	(a1)+,d5		; red
	btst	#2,d5		; 2nd bit set?
	beq.s	nobit1
	moveq	#1,d3
nobit1	andi.b	#%11111000,d5	; mask out lower 3 bits
	lsl.w	#3,d5		; into corresponding bits
	move.w	(a1)+,d2		; green
	btst	#2,d2
	beq.s	nobit2
	addq.b	#1,d3
nobit2	lsr.b	#3,d2		; mask out lower 2 bits
	lsl.w	#3,d2		; and move up
	lsl.w	#8,d2
	or.w	d2,d5
	move.w	(a1)+,d2		; blue
	btst	#2,d2
	beq.s	nobit3
	addq.b	#1,d3
nobit3	lsr.b	#2,d2		; mask out lower 3 bits
	or.b	d2,d5		; now we have rrrrrgggggbbbbbx
	subq.b	#2,d3		; were at least 2 bits set?
	bmi.s	nobit4		; no->...
	bset	#0,d5		; yes, set intensity bit
nobit4	clr.l	d3
	rts
	      


; now we can copy each line from the image, possibly adding a border for each line
; here
; a1 = start of line to copy from
; a3 = current pixel in ql image
; a4 = pc image
; a5 = ql image
; d0 =
; d1 =
; d2 =
; d3 .w = nbr of pixels in x border
; d4 : nbr of lines in border (y dim) | nbr x pixel in ql image
; d5.w = bytes/line (pc)| border colour
; d6 = y | x size of image to copy
; d7 = (loop counter)
; mode 33 is gggggrrr rrbbbbbi

; this is the outer loop - cover each line (row) to treat
oloop33 swap	d6		; d6 = y | x size of image
	move.w	d6,d7		; nbr of pixels to copy
	swap	d6		; back to nbr of lines
	move.l	a1,a2		; point to 1rst byte of this line
	move.w	d3,d0		; any border at left hand side?
	blt.s	no_cpy33		; no, so start converting pixels->...
ilp_cp33	move.w	d5,(a3)+		; copy left border for this line
	dbf	d0,ilp_cp33
no_cpy33	clr.w	d0

; this is the inner loop:  covers each pixel in a line (row)
; bgr
; due to the need to check the 2nd bit of each colour word, this is quite a
; bit longer than the mode32 loop!
iloop33 clr.w	d2
	move.b	1(a2),d0		; green
	andi.b	#%11111100,d0	; get rid of 2 loxest bits
	bclr	#2,d0		; also test & clear bit 2
	beq.s	nobit7		; bit wans set, anyway->...
	moveq	#1,d2		; show that one bit was set
nobit7	lsl.w	#8,d0		; move into upper byte: gggggxxx xxxxxxxx
	move.b	2(a2),d1		; red
	andi.w	#%11111100,d1	; get rid of 2 lowest bits (and upper byte)
	bclr	#2,d1		; also test & clear bit 2
	beq.s	nobit8
	addq.b	#1,d2
nobit8	lsl.w	#3,d1
	or.w	d1,d0		;
	move.b	(a2),d1 	; blue
	bclr	#2,d1		; test & clear
	beq.s	nobit9
	addq.b	#1,d2		; show bit was set
nobit9	lsr.b	#2,d1
	or.b	d1,d0
	subq.b	#2,d2
	bmi.s	nobit10
	bset	#0,d0
nobit10 move.w	d0,(a3)+
	addq.l	#3,a2		; point next bgr
	dbf	d7,iloop33	; treat 1 line
; end of inner loop
	move.w	d3,d0		; any border at right hand side?
	blt.s	no_bord33	; no->...
ilp_cp4 move.w	d5,(a3)+		; copy left border for this line
	dbf	d0,ilp_cp4
no_bord33 swap	d5
	clr.l	d0
	move.w	d5,d0
	sub.l	d0,a1		; point to beginning of next line
	swap	d5
	dbf	d6,oloop33	; do loop
	rts
      
	end
